#!/usr/bin/python
# Copyright (C) 2021 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from dataclasses import dataclass, field
import json
import logging
import os
import subprocess
import sys
import tempfile
from typing import Optional, Dict, List, Tuple, Union
from breezy.commit import PointlessCommit
from breezy.workspace import reset_tree, check_clean_tree
from breezy.workingtree import WorkingTree


class ScriptMadeNoChanges(Exception):
    "Script made no changes."


class ScriptFailed(Exception):
    """Script failed to run."""


class DetailedFailure(Exception):
    """Detailed failure"""

    def __init__(self, result_code, description, details=None):
        self.result_code = result_code
        self.description = description
        self.details = details

    @classmethod
    def from_json(cls, json):
        return cls(
            result_code=json.get('result_code'),
            description=json.get('description'),
            details=json.get('details'))


class ResultFileFormatError(Exception):
    """The result file was invalid."""

    def __init__(self, inner_error):
        self.inner_error = inner_error


@dataclass
class CommandResult(object):

    description: Optional[str] = None
    value: Optional[int] = None
    serialized_context: Optional[str] = None
    context: Dict[str, str] = field(default_factory=dict)
    tags: List[Tuple[str, bytes]] = field(default_factory=list)
    old_revision: Optional[bytes] = None
    new_revision: Optional[bytes] = None
    target_branch_url: Optional[str] = None

    @classmethod
    def from_json(cls, data):
        if 'tags' in data:
            tags = []
            for name, revid in data['tags']:
                tags.append((name, revid.encode('utf-8')))
        else:
            tags = None
        return cls(
            value=data.get('value', None),
            context=data.get('context', {}),
            description=data.get('description'),
            serialized_context=data.get('serialized_context', None),
            target_branch_url=data.get('target-branch-url', None),
            tags=tags)


def script_runner(  # noqa: C901
    local_tree: WorkingTree, script: Union[str, List[str]],
    commit_pending: Optional[bool] = None,
    resume_metadata=None, subpath: str = '', committer: Optional[str] = None,
    extra_env: Optional[Dict[str, str]] = None,
) -> CommandResult:  # noqa: C901
    """Run a script in a tree and commit the result.

    This ignores newly added files.

    Args:
      local_tree: Local tree to run script in
      script: Script to run
      commit_pending: Whether to commit pending changes
        (True, False or None: only commit if there were no commits by the
         script)
    """
    env = dict(os.environ)
    if extra_env:
        env.update(extra_env)
    env['SVP_API'] = '1'
    last_revision = local_tree.last_revision()
    orig_tags = local_tree.branch.tags.get_tag_dict()
    with tempfile.TemporaryDirectory() as td:
        env['SVP_RESULT'] = os.path.join(td, 'result.json')
        if resume_metadata:
            env['SVP_RESUME'] = os.path.join(td, 'resume-metadata.json')
            with open(env['SVP_RESUME'], 'w') as f:
                json.dump(resume_metadata, f)
        p = subprocess.Popen(
            script, cwd=local_tree.abspath(subpath), stdout=subprocess.PIPE,
            shell=isinstance(script, str),
            env=env)
        (description_encoded, err) = p.communicate(b"")
        try:
            with open(env['SVP_RESULT'], 'r') as f:
                try:
                    result_json = json.load(f)
                except json.decoder.JSONDecodeError as e:
                    raise ResultFileFormatError(e)
        except FileNotFoundError:
            result_json = None
        if p.returncode != 0:
            if result_json is not None:
                raise DetailedFailure.from_json(result_json)
            raise ScriptFailed(script, p.returncode)
        if result_json is not None:
            result = CommandResult.from_json(result_json)
        else:
            result = CommandResult()
    if not result.description:
        result.description = description_encoded.decode()
    new_revision = local_tree.last_revision()
    if result.tags is None:
        result.tags = []
        for name, revid in local_tree.branch.tags.get_tag_dict().items():
            if orig_tags.get(name) != revid:
                result.tags.append((name, revid))
    if last_revision == new_revision and commit_pending is None:
        # Automatically commit pending changes if the script did not
        # touch the branch.
        commit_pending = True
    if commit_pending:
        local_tree.smart_add([local_tree.abspath(subpath)])
        try:
            new_revision = local_tree.commit(
                result.description, allow_pointless=False,
                committer=committer)
        except PointlessCommit:
            pass
    if new_revision == last_revision:
        raise ScriptMadeNoChanges()
    result.old_revision = last_revision
    result.new_revision = local_tree.last_revision()
    return result


def main(argv: List[str]) -> Optional[int]:  # noqa: C901
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "command", help="Path to script to run.", type=str,
        nargs='?')
    parser.add_argument(
        "--diff", action="store_true", help="Show diff of generated changes."
    )
    parser.add_argument(
        "--commit-pending",
        help="Commit pending changes after script.",
        choices=["yes", "no", "auto"],
        default=None,
        type=str,
    )
    parser.add_argument(
        "--verify-command", type=str, help="Command to run to verify changes."
    )
    parser.add_argument(
        "--recipe", type=str, help="Recipe to use.")
    args = parser.parse_args(argv)

    if args.recipe:
        from .recipe import Recipe
        recipe = Recipe.from_path(args.recipe)
    else:
        recipe = None

    if args.commit_pending:
        commit_pending = {
            "auto": None, "yes": True, "no": False}[args.commit_pending]
    elif recipe:
        commit_pending = recipe.commit_pending
    else:
        commit_pending = None

    if args.command:
        command = args.command
    elif recipe.command:
        command = recipe.command
    else:
        logging.exception('No command specified.')
        return 1

    local_tree, subpath = WorkingTree.open_containing('.')

    check_clean_tree(local_tree)

    try:
        result = script_runner(
            local_tree, script=command, commit_pending=commit_pending,
            subpath=subpath)

        if result.description:
            logging.info('Succeeded: %s', result.description)

        if args.verify_command:
            try:
                subprocess.check_call(
                    args.verify_command, shell=True,
                    cwd=local_tree.abspath(subpath)
                )
            except subprocess.CalledProcessError:
                logging.exception("Verify command failed.")
                return 1
    except Exception:
        reset_tree(local_tree, subpath)
        raise

    if args.diff:
        from breezy.diff import show_diff_trees
        old_tree = local_tree.revision_tree(result.old_revision)
        new_tree = local_tree.revision_tree(result.new_revision)
        show_diff_trees(
            old_tree,
            new_tree,
            sys.stdout.buffer,
            old_label='old/',
            new_label='new/')
    return 0



#!/usr/bin/python
# Copyright (C) 2021 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from dataclasses import dataclass
from typing import Optional, List
import yaml


@dataclass
class Candidate(object):
    """Candidate."""

    url: str
    branch: Optional[str] = None
    subpath: str = ''

    @classmethod
    def from_yaml(cls, d):
        if isinstance(d, dict):
            return cls(
                url=d.get('url'),
                branch=d.get('branch'),
                subpath=d.get('path'),
                )
        elif isinstance(d, str):
            return cls(url=d)
        else:
            raise TypeError(d)


@dataclass
class CandidateList(object):
    """Candidate list."""

    candidates: List[Candidate]

    def __iter__(self):
        return iter(self.candidates)

    @classmethod
    def from_yaml(cls, d):
        candidates = []
        for entry in d:
            candidates.append(Candidate.from_yaml(entry))
        return cls(candidates=candidates)

    @classmethod
    def from_path(cls, path):
        with open(path, 'r') as f:
            return cls.from_yaml(yaml.full_load(f))



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for uploading packages."""

import silver_platter  # noqa: F401

import datetime
from email.utils import parseaddr
import logging
import os
import subprocess
import sys
import tempfile
from typing import Optional

from debian.changelog import Version

from debmutate.changelog import (
    ChangelogEditor,
    ChangelogParseError,
    changeblock_ensure_first_line,
    gbp_dch,
)
from debmutate.control import ControlEditor

from breezy import gpg
from breezy.config import extract_email_address
from breezy.errors import NoSuchTag, PermissionDenied
from breezy.commit import NullCommitReporter
from breezy.revision import NULL_REVISION
from breezy.plugins.debian.builder import BuildFailedError
from breezy.plugins.debian.cmds import _build_helper
from breezy.plugins.debian.import_dsc import (
    DistributionBranch,
)
from breezy.plugins.debian.release import (
    release,
)
from breezy.plugins.debian.util import (
    changelog_find_previous_upload,
    dput_changes,
    find_changelog,
    MissingChangelogError,
    NoPreviousUpload,
)
from breezy.plugins.debian.upstream import MissingUpstreamTarball

from breezy.workingtree import WorkingTree

from debian.changelog import get_maintainer

from . import (
    apt_get_source_package,
    source_package_vcs,
    split_vcs_url,
    Workspace,
    DEFAULT_BUILDER,
    select_probers,
    NoSuchPackage,
)
from ..utils import (
    open_branch,
    BranchUnavailable,
    BranchMissing,
    BranchUnsupported,
    BranchRateLimited,
)


def debsign(path, keyid=None):
    (bd, changes_file) = os.path.split(path)
    args = ["debsign"]
    if keyid:
        args.append("-k%s" % keyid)
    args.append(changes_file)
    subprocess.check_call(args, cwd=bd)


class LastUploadMoreRecent(Exception):
    """Last version in archive is newer than vcs version."""

    def __init__(self, archive_version, vcs_version):
        self.archive_version = archive_version
        self.vcs_version = vcs_version
        super(LastUploadMoreRecent, self).__init__(
            "last upload (%s) is more recent than vcs (%s)"
            % (archive_version, vcs_version)
        )


class NoUnuploadedChanges(Exception):
    """Indicates there are no unuploaded changes for a package."""

    def __init__(self, archive_version):
        self.archive_version = archive_version
        super(NoUnuploadedChanges, self).__init__(
            "nothing to upload, latest version is in archive: %s" %
            archive_version
        )


class NoUnreleasedChanges(Exception):
    """Indicates there are no unreleased changes for a package."""

    def __init__(self, version):
        self.version = version
        super(NoUnreleasedChanges, self).__init__(
            "nothing to upload, latest version in vcs is not unreleased: %s" %
            version
        )


class RecentCommits(Exception):
    """Indicates there are too recent commits for a package."""

    def __init__(self, commit_age, min_commit_age):
        self.commit_age = commit_age
        self.min_commit_age = min_commit_age
        super(RecentCommits, self).__init__(
            "Last commit is only %d days old (< %d)"
            % (self.commit_age, self.min_commit_age)
        )


class CommitterNotAllowed(Exception):
    """Specified committer is not allowed."""

    def __init__(self, committer, allowed_committers):
        self.committer = committer
        self.allowed_committers = allowed_committers
        super(CommitterNotAllowed, self).__init__(
            "Committer %s not in allowed committers: %r"
            % (self.committer, self.allowed_committers)
        )


class LastReleaseRevisionNotFound(Exception):
    """The revision for the last uploaded release can't be found."""

    def __init__(self, package, version):
        self.package = package
        self.version = version
        super(LastReleaseRevisionNotFound, self).__init__(
            "Unable to find revision matching version %r for %s" %
            (version, package)
        )


def check_revision(rev, min_commit_age, allowed_committers):
    """Check whether a revision can be included in an upload.

    Args:
      rev: revision to check
      min_commit_age: Minimum age for revisions
      allowed_committers: List of allowed committers
    Raises:
      RecentCommits: When there are commits younger than min_commit_age
    """
    # TODO(jelmer): deal with timezone
    if min_commit_age is not None:
        commit_time = datetime.datetime.fromtimestamp(rev.timestamp)
        time_delta = datetime.datetime.now() - commit_time
        if time_delta.days < min_commit_age:
            raise RecentCommits(time_delta.days, min_commit_age)
    # TODO(jelmer): Allow tag to prevent automatic uploads
    committer_email = extract_email_address(rev.committer)
    if allowed_committers and committer_email not in allowed_committers:
        raise CommitterNotAllowed(committer_email, allowed_committers)


def find_last_release_revid(branch, version):
    db = DistributionBranch(branch, None)
    return db.revid_of_version(version)


def get_maintainer_keys(context):
    for key in context.keylist(
            source="/usr/share/keyrings/debian-keyring.gpg"):
        yield key.fpr
        for subkey in key.subkeys:
            yield subkey.keyid


class GbpDchFailed(Exception):
    """gbp dch failed to run"""


def prepare_upload_package(  # noqa: C901
    local_tree: WorkingTree,
    subpath: str,
    pkg: str,
    last_uploaded_version: Optional[Version],
    builder: str,
    gpg_strategy: Optional[gpg.GPGStrategy] = None,
    min_commit_age=None,
    allowed_committers=None,
):
    debian_path = os.path.join(subpath, "debian")
    try:
        from lintian_brush.detect_gbp_dch import guess_update_changelog
    except ImportError:
        run_gbp_dch = True   # Let's just try
    else:
        cl_behaviour = guess_update_changelog(local_tree, debian_path)
        run_gbp_dch = (
            cl_behaviour is None or not cl_behaviour.update_changelog)
    if run_gbp_dch:
        try:
            gbp_dch(local_tree.abspath(subpath))
        except subprocess.CalledProcessError:
            # TODO(jelmer): gbp dch sometimes fails when there is no existing
            # open changelog entry; it fails invoking
            # "dpkg --lt None <old-version>"
            raise GbpDchFailed()
        local_tree.commit(
            specific_files=[os.path.join(debian_path, 'changelog')],
            message='update changelog\n\nGbp-Dch: Ignore')
    cl, top_level = find_changelog(local_tree, merge=False, max_blocks=None)
    if (last_uploaded_version is not None
            and cl.version == last_uploaded_version):
        raise NoUnuploadedChanges(cl.version)
    try:
        previous_version_in_branch = changelog_find_previous_upload(cl)
    except NoPreviousUpload:
        pass
    else:
        if (last_uploaded_version is not None and
                last_uploaded_version > previous_version_in_branch):
            raise LastUploadMoreRecent(
                last_uploaded_version, previous_version_in_branch)

    if last_uploaded_version is not None:
        logging.info("Checking revisions since %s", last_uploaded_version)
    with local_tree.lock_read():
        if last_uploaded_version is not None:
            try:
                last_release_revid = find_last_release_revid(
                    local_tree.branch, last_uploaded_version
                )
            except NoSuchTag:
                raise LastReleaseRevisionNotFound(pkg, last_uploaded_version)
        else:
            last_release_revid = NULL_REVISION
        graph = local_tree.branch.repository.get_graph()
        revids = list(
            graph.iter_lefthand_ancestry(
                local_tree.branch.last_revision(), [last_release_revid]
            )
        )
        if not revids:
            logging.info("No pending changes")
            return
        if gpg_strategy:
            logging.info("Verifying GPG signatures...")
            count, result, all_verifiables = gpg.bulk_verify_signatures(
                local_tree.branch.repository, revids, gpg_strategy
            )
            for revid, code, key in result:
                if code != gpg.SIGNATURE_VALID:
                    raise Exception(
                        "No valid GPG signature on %r: %d" % (revid, code))
        for revid, rev in local_tree.branch.repository.iter_revisions(revids):
            if rev is not None:
                check_revision(rev, min_commit_age, allowed_committers)

        if cl.distributions != "UNRELEASED":
            raise NoUnreleasedChanges(cl.version)
    qa_upload = False
    team_upload = False
    control_path = local_tree.abspath(os.path.join(debian_path, "control"))
    with ControlEditor(control_path) as e:
        maintainer = parseaddr(e.source["Maintainer"])
        if maintainer[1] == "packages@qa.debian.org":
            qa_upload = True
        # TODO(jelmer): Check whether this is a team upload
        # TODO(jelmer): determine whether this is a NMU upload
    if qa_upload or team_upload:
        changelog_path = local_tree.abspath(
            os.path.join(debian_path, "changelog"))
        with ChangelogEditor(changelog_path) as e:
            if qa_upload:
                changeblock_ensure_first_line(e[0], "QA upload.")
                message = "Mention QA Upload."
            elif team_upload:
                changeblock_ensure_first_line(e[0], "Team upload.")
                message = "Mention Team Upload."
            else:
                message = None
        if message is not None:
            local_tree.commit(
                specific_files=[os.path.join(debian_path, "changelog")],
                message=message,
                allow_pointless=False,
                reporter=NullCommitReporter(),
            )
    tag_name = release(local_tree, subpath)
    target_dir = tempfile.mkdtemp()
    if last_uploaded_version is not None:
        builder = builder.replace(
            "${LAST_VERSION}", str(last_uploaded_version))
    target_changes = _build_helper(
        local_tree, subpath, local_tree.branch, target_dir, builder=builder
    )
    debsign(target_changes['source'])
    return target_changes['source'], tag_name


def select_apt_packages(package_names, maintainer):
    packages = []
    import apt_pkg

    apt_pkg.init()
    sources = apt_pkg.SourceRecords()
    while sources.step():
        if maintainer:
            fullname, email = parseaddr(sources.maintainer)
            if email not in maintainer:
                continue

        if package_names and sources.package not in package_names:
            continue

        packages.append(sources.package)

    return packages


class PackageProcessingFailure(Exception):

    def __init__(self, reason):
        self.reason = reason


class PackageIgnored(Exception):

    def __init__(self, reason):
        self.reason = reason


def check_git_commits(vcslog, min_commit_age, allowed_committers):
    class GitRevision:

        @property
        def committer(self):
            return self.headers.get('Committer') or self.headers.get('Author')

        @property
        def timestamp(self):
            datestr = self.headers.get('Date')
            dt = datetime.datetime.strptime(
                datestr.strip(), "%a %b %d %H:%M:%S %Y %z")
            return dt.timestamp()

        def __init__(self, commit_id, headers, message):
            self.commit_id = commit_id
            self.headers = headers
            self.message = message

        @classmethod
        def from_lines(cls, lines):
            commit_id = None
            message = []
            headers = {}
            for i, line in enumerate(lines):
                if line.startswith('commit '):
                    commit_id = line[len('commit '):]
                elif line == '':
                    message = lines[i+1:]
                    break
                else:
                    name, value = line.split(': ')
                    headers[name] = value
            return cls(commit_id, headers, message)

    last_commit_ts = None
    lines = []
    for line in vcslog.splitlines():
        if line == '' and lines[-1][0].isspace():
            gitrev = GitRevision.from_lines(lines)
            if last_commit_ts is None:
                last_commit_ts = gitrev.timestamp
            check_revision(
                gitrev, min_commit_age,
                allowed_committers)
            lines = []
        else:
            lines.append(line)
    if lines:
        gitrev = GitRevision.from_lines(lines)
        if last_commit_ts is None:
            last_commit_ts = gitrev.timestamp
        check_revision(gitrev, min_commit_age, allowed_committers)
    return last_commit_ts


def process_package(
        package,
        builder: str, exclude=None, autopkgtest_only: bool = False,
        gpg_verification: bool = False,
        acceptable_keys=None, debug: bool = False, dry_run: bool = False,
        diff: bool = False, min_commit_age=None, allowed_committers=None,
        vcs_type=None, vcs_url=None, source_name=None,
        archive_version=None):
    if exclude is None:
        exclude = set()
    logging.info("Processing %s", package)
    # Can't use open_packaging_branch here, since we want to use pkg_source
    # later on.
    if "/" not in package:
        try:
            pkg_source = apt_get_source_package(package)
        except NoSuchPackage:
            logging.info("%s: package not found in apt", package)
            raise PackageProcessingFailure('not-in-apt')
        if vcs_type is None or vcs_url is None:
            try:
                vcs_type, vcs_url = source_package_vcs(pkg_source)
            except KeyError:
                logging.info(
                    "%s: no declared vcs location, skipping",
                    pkg_source["Package"]
                )
                raise PackageProcessingFailure('not-in-vcs')
        if source_name is None:
            source_name = pkg_source["Package"]
        if source_name in exclude:
            raise PackageIgnored('excluded')
        if archive_version is None:
            archive_version = pkg_source["Version"]
        has_testsuite = "Testsuite" in pkg_source
    else:
        if vcs_url is None:
            vcs_url = package
        has_testsuite = None
    (location, branch_name, subpath) = split_vcs_url(vcs_url)
    if subpath is None:
        subpath = ""
    probers = select_probers(vcs_type)
    try:
        main_branch = open_branch(
            location, probers=probers, name=branch_name)
    except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
        if debug:
            logging.exception("%s: %s", vcs_url, e)
        else:
            logging.info(
                "%s: branch unavailable: %s", vcs_url, e)
        raise PackageProcessingFailure('vcs-inaccessible')
    with Workspace(main_branch) as ws:
        if source_name is None:
            with ControlEditor(
                ws.local_tree.abspath(
                    os.path.join(subpath, "debian/control"))
            ) as ce:
                source_name = ce.source["Source"]
                try:
                    pkg_source = apt_get_source_package(source_name)
                except NoSuchPackage:
                    logging.info("%s: package not found in apt", package)
                    raise PackageProcessingFailure('not-in-apt')
                archive_version = pkg_source['Version']
            has_testsuite = "Testsuite" in ce.source
        if source_name in exclude:
            raise PackageIgnored('excluded')
        if (
            autopkgtest_only
            and not has_testsuite
            and not ws.local_tree.has_filename(
                os.path.join(subpath, "debian/tests/control")
            )
        ):
            logging.info(
                "%s: Skipping, package has no autopkgtest.", source_name)
            raise PackageIgnored('no-autopkgtest')
        branch_config = ws.local_tree.branch.get_config_stack()
        if gpg_verification:
            gpg_strategy = gpg.GPGStrategy(branch_config)
            if acceptable_keys:
                acceptable_keys = acceptable_keys
            else:
                acceptable_keys = list(
                    get_maintainer_keys(gpg_strategy.context))
            gpg_strategy.set_acceptable_keys(",".join(acceptable_keys))
        else:
            gpg_strategy = None

        try:
            target_changes, tag_name = prepare_upload_package(
                ws.local_tree,
                subpath,
                source_name,
                archive_version,
                builder=builder,
                gpg_strategy=gpg_strategy,
                min_commit_age=min_commit_age,
                allowed_committers=allowed_committers,
            )
        except GbpDchFailed as e:
            logging.warn("%s: 'gbp dch' failed to run: %s", source_name, e)
            raise PackageProcessingFailure('gbp-dch-failed')
        except MissingUpstreamTarball as e:
            logging.warning(
                "%s: missing upstream tarball: %s", source_name, e)
            raise PackageProcessingFailure('missing-upstream-tarball')
        except BranchRateLimited as e:
            logging.warning(
                '%s: rate limited by server (retrying after %s)',
                source_name, e.retry_after)
            raise PackageProcessingFailure('rate-limited')
        except CommitterNotAllowed as e:
            logging.warn(
                "%s: committer %s not in allowed list: %r",
                source_name,
                e.committer,
                e.allowed_committers,
            )
            raise PackageIgnored('committer-not-allowed')
        except BuildFailedError as e:
            logging.warn("%s: package failed to build: %s", source_name, e)
            raise PackageProcessingFailure('build-failed')
        except LastReleaseRevisionNotFound as e:
            logging.warn(
                "%s: Unable to find revision matching last release "
                "%s, skipping.",
                source_name,
                e.version,
            )
            raise PackageProcessingFailure('last-release-missing')
        except LastUploadMoreRecent as e:
            logging.warn(
                "%s: Last upload (%s) was more recent than VCS (%s)",
                source_name,
                e.archive_version,
                e.vcs_version,
            )
            raise PackageProcessingFailure('last-upload-not-in-vcs')
        except ChangelogParseError as e:
            logging.info("%s: Error parsing changelog: %s", source_name, e)
            raise PackageProcessingFailure('changelog-parse-error')
        except MissingChangelogError:
            logging.info("%s: No changelog found, skipping.", source_name)
            raise PackageProcessingFailure('missing-changelog')
        except RecentCommits as e:
            logging.info(
                "%s: Recent commits (%d days), skipping.",
                source_name, e.commit_age
            )
            raise PackageIgnored('recent-commits')
        except NoUnuploadedChanges as e:
            logging.info("%s: No unuploaded changes (%s), skipping.",
                         source_name, e.archive_version)
            raise PackageIgnored('no-unuploaded-changes')
        except NoUnreleasedChanges:
            logging.info(
                "%s: No unreleased changes, skipping.", source_name)
            raise PackageIgnored('no-unreleased-changes')

        tags = []
        if tag_name is not None:
            logging.info("Pushing tag %s", tag_name)
            tags.append(tag_name)
        try:
            ws.push(dry_run=dry_run, tags=tags)
        except PermissionDenied:
            logging.info(
                "%s: Permission denied pushing to branch, skipping.",
                source_name
            )
            raise PackageProcessingFailure('vcs-permission-denied')
        if not dry_run:
            dput_changes(target_changes)
        if diff:
            sys.stdout.flush()
            ws.show_diff(sys.stdout.buffer)
            sys.stdout.buffer.flush()


def vcswatch_prescan_package(
        package, vw, exclude=None, min_commit_age=None,
        allowed_committers=None):
    vcs_url = vw["url"]
    vcs_type = vw["vcs"]
    source_name = vw["package"]
    if source_name in exclude:
        raise PackageIgnored('excluded')
    if vcs_url is None or vcs_type is None:
        raise PackageProcessingFailure('not-in-vcs')
    # TODO(jelmer): check autopkgtest_only ?
    # from debian.deb822 import Deb822
    # pkg_source = Deb822(vw["controlfile"])
    # has_testsuite = "Testsuite" in pkg_source
    if vw["commits"] == 0:
        raise PackageIgnored('no-unuploaded-changes')
    if vw['status'] == 'ERROR':
        logging.warning(
            'vcswatch: unable to access %s: %s', vw['package'], vw['error'])
        raise PackageProcessingFailure('vcs-inaccessible')
    logging.debug("vcswatch last scanned at: %s", vw["last_scan"])
    if vcs_type == 'Git' and vw["vcslog"] is not None:
        try:
            return check_git_commits(
                vw["vcslog"], min_commit_age=min_commit_age,
                allowed_committers=allowed_committers)
        except CommitterNotAllowed as e:
            logging.warn(
                "%s: committer %s not in allowed list: %r",
                source_name,
                e.committer,
                e.allowed_committers,
            )
            raise PackageIgnored('committer-not-allowed')
        except RecentCommits as e:
            logging.info(
                "%s: Recent commits (%d days), skipping.",
                source_name, e.commit_age
            )
            raise PackageIgnored('recent-commits')
    return None


def vcswatch_prescan_packages(
        packages, inc_stats, exclude=None, min_commit_age=None,
        allowed_committers=None):
    logging.info('Using vcswatch to prescan %d packages', len(packages))
    from .. import version_string
    from urllib.request import Request, urlopen
    import json
    import gzip
    url = "https://qa.debian.org/data/vcswatch/vcswatch.json.gz"
    request = Request(url, headers={
        'User-Agent': "silver-platter/%s" % version_string})
    vcswatch = {
        p['package']: p
        for p in json.load(gzip.GzipFile(fileobj=urlopen(request)))}
    by_ts = {}
    failures = 0
    for package in packages:
        try:
            vw = vcswatch[package]
        except KeyError:
            continue
        try:
            ts = vcswatch_prescan_package(
                    package, vw,
                    exclude=exclude, min_commit_age=min_commit_age,
                    allowed_committers=allowed_committers)
        except PackageProcessingFailure as e:
            inc_stats(e.reason)
            failures += 1
        except PackageIgnored as e:
            inc_stats(e.reason)
        else:
            by_ts[package] = ts
    return ([k for (k, v) in
             sorted(by_ts.items(), key=lambda k: k[1] or 0, reverse=True)],
            failures,
            vcswatch)


def main(argv):  # noqa: C901
    import argparse

    parser = argparse.ArgumentParser(prog="upload-pending-commits")
    parser.add_argument("packages", nargs="*")
    parser.add_argument(
        "--dry-run", action="store_true", help="Dry run changes.")
    parser.add_argument(
        "--acceptable-keys",
        help="List of acceptable GPG keys",
        action="append",
        default=[],
        type=str,
    )
    parser.add_argument(
        "--gpg-verification",
        help="Verify GPG signatures on commits",
        action="store_true",
    )
    parser.add_argument(
        "--min-commit-age",
        help="Minimum age of the last commit, in days",
        type=int,
        default=0,
    )
    parser.add_argument("--diff", action="store_true", help="Show diff.")
    parser.add_argument(
        "--builder",
        type=str,
        help="Build command",
        default=(
            DEFAULT_BUILDER + " --source --source-only-changes "
            "--debbuildopt=-v${LAST_VERSION}"
        ),
    )
    parser.add_argument(
        "--maintainer",
        type=str,
        action="append",
        help="Select all packages maintained by specified maintainer.",
    )
    parser.add_argument(
        "--vcswatch",
        action="store_true",
        default=False,
        help="Use vcswatch to determine what packages need uploading."
    )
    parser.add_argument(
        "--exclude", type=str, action="append", default=[],
        help="Ignore source package"
    )
    parser.add_argument(
        "--autopkgtest-only",
        action="store_true",
        help="Only process packages with autopkgtests.",
    )
    parser.add_argument(
        "--allowed-committer",
        type=str,
        action="append",
        help="Require that all new commits are from specified committers",
    )
    parser.add_argument(
        "--debug",
        action="store_true")
    parser.add_argument(
        "--shuffle",
        action="store_true",
        help="Randomize order packages are processed in.")

    args = parser.parse_args(argv)

    ret = 0

    if not args.packages and not args.maintainer:
        (name, email) = get_maintainer()
        if email:
            logging.info("Processing packages maintained by %s", email)
            args.maintainer = [email]
        else:
            parser.print_usage()
            sys.exit(1)

    if not args.vcswatch:
        logging.info(
            "Use --vcswatch to only process packages for which "
            "vcswatch found pending commits."
        )
    if args.maintainer:
        packages = select_apt_packages(args.packages, args.maintainer)
    else:
        packages = args.packages

    if not packages:
        logging.info("No packages found.")
        parser.print_usage()
        sys.exit(1)

    if args.shuffle:
        import random
        random.shuffle(packages)

    stats = {}

    def inc_stats(result):
        stats.setdefault(result, 0)
        stats[result] += 1

    if args.vcswatch:
        packages, failures, extra_data = vcswatch_prescan_packages(
            packages, inc_stats, exclude=args.exclude,
            min_commit_age=args.min_commit_age,
            allowed_committers=args.allowed_committer)
        if failures > 0:
            ret = 1
    else:
        extra_data = {}

    if len(packages) > 1:
        logging.info("Uploading packages: %s", ", ".join(packages))

    for package in packages:
        vcs_type = extra_data.get(package, {}).get('vcs_type')
        vcs_url = extra_data.get(package, {}).get('vcs_url')
        archive_version = extra_data.get(package, {}).get('archive_version')
        source_name = extra_data.get(package, {}).get('package')

        try:
            process_package(
                package, builder=args.builder, exclude=args.exclude,
                autopkgtest_only=args.autopkgtest_only,
                gpg_verification=args.gpg_verification,
                acceptable_keys=args.acceptable_keys,
                debug=args.debug, dry_run=args.dry_run,
                diff=args.diff, min_commit_age=args.min_commit_age,
                allowed_committers=args.allowed_committer,
                vcs_type=vcs_type, vcs_url=vcs_url,
                archive_version=archive_version,
                source_name=source_name)
        except PackageProcessingFailure as e:
            inc_stats(e.reason)
            ret = 1
        except PackageIgnored as e:
            inc_stats(e.reason)

    if len(packages) > 1:
        logging.info('Results:')
        for error, c in stats.items():
            logging.info('  %s: %d', error, c)

    return ret


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/python
# Copyright (C) 2021 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from dataclasses import dataclass
from jinja2 import Template
from typing import Optional, Dict, Union, List
import yaml


@dataclass
class Recipe(object):
    """Recipe to use."""

    name: str
    command: Union[str, List[str]]
    merge_request_description_template: Dict[Optional[str], Template]
    merge_request_commit_message_template: Template
    resume: bool = False
    commit_pending: Optional[bool] = True
    propose_threshold: Optional[int] = None

    @classmethod
    def from_yaml(cls, d):
        merge_request = d.get('merge-request', {})
        if merge_request:
            description = merge_request.get('description', {})
            if isinstance(description, dict):
                merge_request_description_template = description
            else:
                merge_request_description_template = {None: description}
            merge_request_commit_message_template = merge_request.get(
                'commit-message')
            propose_threshold = merge_request.get('propose-threshold')
        else:
            merge_request_description_template = {}
            merge_request_commit_message_template = None
            propose_threshold = None
        return cls(
            name=d.get('name'),
            command=d.get('command'),
            resume=d.get('resume', False),
            commit_pending=d.get('commit-pending'),
            merge_request_description_template=(
                merge_request_description_template),
            merge_request_commit_message_template=(
                merge_request_commit_message_template),
            propose_threshold=propose_threshold)

    def render_merge_request_commit_message(self, context):
        template = self.merge_request_commit_message_template
        if template:
            return Template(template).render(context)
        return None

    def render_merge_request_description(self, description_format, context):
        template = self.merge_request_description_template.get(
                description_format)
        if template is None:
            try:
                template = self.merge_request_description_template[None]
            except KeyError:
                return None
        return Template(template).render(context)

    @classmethod
    def from_path(cls, path):
        with open(path, 'r') as f:
            return cls.from_yaml(yaml.full_load(f))



#!/usr/bin/python
# Copyright (C) 2018-2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import logging
from typing import (
    Optional,
    Callable,
    List,
    Union,
    Dict,
    BinaryIO,
    Any,
    Tuple,
)

from breezy.branch import Branch
from breezy.controldir import ControlDir
try:
    from breezy.controldir import NoColocatedBranchSupport
except ImportError:  # breezy < 3.3
    from breezy.errors import NoColocatedBranchSupport
from breezy.tree import Tree
from breezy.workingtree import WorkingTree
from breezy.diff import show_diff_trees
from breezy.errors import (
    DivergedBranches,
    NotBranchError,
)
from .proposal import (
    get_forge,
    Forge,
    MergeProposal,
    UnsupportedForge,
)

from breezy.transport.local import LocalTransport

from .publish import (
    propose_changes,
    push_changes,
    push_derived_changes,
    publish_changes as _publish_changes,
    PublishResult,
)
from .utils import (
    create_temp_sprout,
    full_branch_url,
)


__all__ = [
    "Workspace",
]


logger = logging.getLogger(__name__)


def fetch_colocated(controldir: ControlDir, from_controldir: ControlDir,
                    additional_colocated_branches: Dict[str, str]):
    logger.debug(
        "Fetching colocated branches: %r",
        additional_colocated_branches,
    )
    for (from_branch_name,
         to_branch_name) in additional_colocated_branches.items():
        try:
            remote_colo_branch = from_controldir.open_branch(
                name=from_branch_name
            )
        except (NotBranchError, NoColocatedBranchSupport):
            continue
        controldir.push_branch(
            name=to_branch_name, source=remote_colo_branch, overwrite=True
        )


class Workspace(object):
    """Workspace for creating changes to a branch.

    Args:
        main_branch: The upstream branch
        resume_branch: Optional in-progress branch that we previously made
            changes on, and should ideally continue from.
        resume_branch_additional_colocated_branches:
            Additional list of colocated branches to fetch
        cached_branch: Branch to copy revisions from, if possible.
        local_tree: The tree the user can work in
    """

    _destroy: Optional[Callable[[], None]]
    local_tree: WorkingTree
    main_branch_revid: Optional[bytes]
    main_colo_revid: Dict[Optional[str], bytes]
    additional_colocated_branches: Dict[str, str]
    resume_branch_additional_colocated_branches: Optional[Dict[str, str]]

    @classmethod
    def from_url(cls, url, **kwargs):
        return cls(main_branch=Branch.open(url), **kwargs)

    def __init__(
        self,
        main_branch: Branch,
        resume_branch: Optional[Branch] = None,
        cached_branch: Optional[Branch] = None,
        additional_colocated_branches:
            Optional[Union[List[str], Dict[str, str]]] = None,
        resume_branch_additional_colocated_branches:
            Optional[Union[List[str], Dict[str, str]]] = None,
        dir: Optional[str] = None,
        path: Optional[str] = None,
    ) -> None:
        self.main_branch = main_branch
        self.main_branch_revid = None
        self.refreshed = False
        self.cached_branch = cached_branch
        self.resume_branch = resume_branch
        if additional_colocated_branches is None:
            additional_colocated_branches = {}
        elif isinstance(additional_colocated_branches, list):
            additional_colocated_branches = {
                k: k for k in additional_colocated_branches}
        self.additional_colocated_branches = additional_colocated_branches
        if isinstance(resume_branch_additional_colocated_branches, list):
            resume_branch_additional_colocated_branches = {
                k: k for k in resume_branch_additional_colocated_branches}
        self.resume_branch_additional_colocated_branches = (
            resume_branch_additional_colocated_branches)
        self._destroy = None
        self._dir = dir
        self._path = path

    @property
    def path(self):
        return self.local_tree.abspath('.')

    def __str__(self):
        if self._path is None:
            return "Workspace for %s" % full_branch_url(self.main_branch)
        else:
            return "Workspace for %s at %s" % (
                full_branch_url(self.main_branch),
                self._path,
            )

    def __repr__(self):
        return (
            "%s(%r, resume_branch=%r, cached_branch=%r, "
            "additional_colocated_branches=%r, "
            "resume_branch_additional_colocated_branches=%r, dir=%r, path=%r)"
            % (
                type(self).__name__,
                self.main_branch,
                self.resume_branch,
                self.cached_branch,
                self.additional_colocated_branches,
                self.resume_branch_additional_colocated_branches,
                self._dir,
                self._path,
            )
        )

    def _inverse_additional_colocated_branches(self):
        return {
            to_name: from_name
            for from_name, to_name in
            self.additional_colocated_branches.items()}

    def __enter__(self) -> Any:
        for (sprout_base, sprout_coloc) in [
                (self.cached_branch, self.additional_colocated_branches),
                (self.resume_branch,
                    self.resume_branch_additional_colocated_branches),
                (self.main_branch, self.additional_colocated_branches)]:
            if sprout_base:
                break
        else:
            raise ValueError('main branch needs to be specified')
        logger.debug("Creating sprout from %r", sprout_base)
        self.local_tree, self._destroy = create_temp_sprout(
            sprout_base,
            sprout_coloc,
            dir=self._dir,
            path=self._path,
        )
        self.main_branch_revid = self.main_branch.last_revision()
        self.main_colo_revid = {}
        for from_name, to_name in self.additional_colocated_branches.items():
            try:
                branch = self.main_branch.controldir.open_branch(
                    name=from_name)  # type: ignore
            except (NotBranchError, NoColocatedBranchSupport):
                continue
            self.main_colo_revid[to_name] = branch.last_revision()

        self.refreshed = False
        if self.cached_branch:
            logger.debug(
                "Pulling in missing revisions from resume/main branch %r",
                self.resume_branch or self.main_branch,
            )
            self.local_tree.pull(
                self.resume_branch or self.main_branch, overwrite=True
            )
        # At this point, we're either on the tip of the main branch or the tip
        # of the resume branch
        if self.resume_branch:
            # If there's a resume branch at play, make sure it's derived from
            # the main branch *or* reset back to the main branch.
            logger.debug(
                "Pulling in missing revisions from main branch %r",
                self.main_branch
            )
            try:
                self.local_tree.pull(self.main_branch, overwrite=False)
            except DivergedBranches:
                logger.info("restarting branch")
                self.refreshed = True
                self.local_tree.pull(self.main_branch, overwrite=True)
                fetch_colocated(
                    self.local_tree.branch.controldir,
                    self.main_branch.controldir,
                    self.additional_colocated_branches)
            else:
                fetch_colocated(
                    self.local_tree.branch.controldir,
                    self.main_branch.controldir,
                    self.additional_colocated_branches)
                if self.resume_branch_additional_colocated_branches:
                    fetch_colocated(
                        self.local_tree.branch.controldir,
                        self.resume_branch.controldir,
                        self.resume_branch_additional_colocated_branches)
                    self.additional_colocated_branches.update(
                        self.resume_branch_additional_colocated_branches)
        else:
            fetch_colocated(
                self.local_tree.branch.controldir,
                self.main_branch.controldir,
                self.additional_colocated_branches)
        self.base_revid = self.local_tree.last_revision()
        return self

    def defer_destroy(self) -> Optional[Callable[[], None]]:
        ret = self._destroy
        self._destroy = None
        return ret

    def changes_since_main(self) -> bool:
        return self.local_tree.branch.last_revision() != self.main_branch_revid

    def changes_since_base(self) -> bool:
        return self.base_revid != self.local_tree.branch.last_revision()

    def any_branch_changes(self):
        """Have any branch changes at all been made?

        Includes changes that already existed in the resume branch.
        """
        for name, br, r in self.result_branches():
            if br != r:
                return True
        return False

    def result_branches(self) -> List[
            Tuple[Optional[str], Optional[bytes], Optional[bytes]]]:
        branches = [
            (self.main_branch.name, self.main_branch_revid,  # type: ignore
             self.local_tree.last_revision())]
        for from_name, to_name in self.additional_colocated_branches.items():
            to_revision: Optional[bytes]
            try:
                to_branch = self.local_tree.controldir.open_branch(
                    name=to_name)
            except NoColocatedBranchSupport:
                continue
            except NotBranchError:
                to_revision = None
            else:
                to_revision = to_branch.last_revision()
            from_revision = self.main_colo_revid.get(from_name)
            if from_revision is None and to_revision is None:
                continue
            branches.append((from_name, from_revision, to_revision))
        return branches

    def push(
        self,
        forge: Optional[Forge] = None,
        dry_run: bool = False,
        tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
        stop_revision: Optional[bytes] = None,
    ) -> None:
        if forge is None:
            try:
                forge = get_forge(self.main_branch)
            except UnsupportedForge:
                if not isinstance(
                        self.main_branch.control_transport, LocalTransport):
                    logging.warning(
                        'Unable to find forge for %s to determine push url, '
                        'trying anyway.', self.main_branch.user_url)
                forge = None
        return push_changes(
            self.local_tree.branch,
            self.main_branch,
            forge=forge,
            additional_colocated_branches=(
                self._inverse_additional_colocated_branches()),
            dry_run=dry_run,
            tags=tags,
            stop_revision=stop_revision,
        )

    def propose(
        self,
        name: str,
        description: str,
        forge: Optional[Forge] = None,
        existing_proposal: Optional[MergeProposal] = None,
        overwrite_existing: Optional[bool] = None,
        labels: Optional[List[str]] = None,
        dry_run: bool = False,
        commit_message: Optional[str] = None,
        reviewers: Optional[List[str]] = None,
        tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
        owner: Optional[str] = None,
        allow_collaboration: bool = False,
        stop_revision: Optional[bytes] = None,
    ) -> Tuple[MergeProposal, bool]:
        if forge is None:
            forge = get_forge(self.main_branch)
        return propose_changes(
            self.local_tree.branch,
            self.main_branch,
            forge=forge,
            name=name,
            mp_description=description,
            resume_branch=self.resume_branch,
            resume_proposal=existing_proposal,
            overwrite_existing=(overwrite_existing or False),
            labels=labels,
            dry_run=dry_run,
            commit_message=commit_message,
            reviewers=reviewers,
            owner=owner,
            additional_colocated_branches=(
                self._inverse_additional_colocated_branches()),
            tags=tags,
            allow_collaboration=allow_collaboration,
            stop_revision=stop_revision,
        )

    def push_derived(
        self,
        name: str,
        forge: Optional[Forge] = None,
        overwrite_existing: Optional[bool] = False,
        owner: Optional[str] = None,
        tags: Optional[Union[Dict[str, bytes], List[str]]] = None,
        stop_revision: Optional[bytes] = None,
    ) -> Tuple[Branch, str]:
        """Push a derived branch.

        Args:
          name: Branch name
          forge: Optional forge to use
          overwrite_existing: Whether to overwrite an existing branch
          tags: Tags list to push
          owner: Owner name
        Returns:
          tuple with remote_branch and public_branch_url
        """
        if forge is None:
            forge = get_forge(self.main_branch)
        return push_derived_changes(
            self.local_tree.branch,
            self.main_branch,
            forge,
            name,
            overwrite_existing=overwrite_existing,
            owner=owner,
            tags=tags,
            stop_revision=stop_revision,
        )

    def publish_changes(self, *args, **kwargs) -> PublishResult:
        """Publish a set of changes."""
        return _publish_changes(
            self.local_tree.branch,
            self.main_branch,
            self.resume_branch,
            *args,
            **kwargs
        )

    def base_tree(self) -> Tree:
        return self.local_tree.branch.repository.revision_tree(self.base_revid)

    def show_diff(
        self, outf: BinaryIO, old_label: str = "old/", new_label: str = "new/"
    ) -> None:
        base_tree = self.base_tree()
        show_diff_trees(
            base_tree,
            self.local_tree.basis_tree(),
            outf,
            old_label=old_label,
            new_label=new_label,
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._destroy:
            self._destroy()
            self._destroy = None
        return False
